// Hakell frontend for Clean, 26.09.2003

implementation module parse

import StdEnv, StdStrictLists, StdListExtensions, StdMaybe//, Directory
//import RWSDebug
import hashtable, containers, genericsupport, StdCompare, syntax, utilities, predef, containers, parse_original

/*
The structure of this module:

* utility functions
* the scanner phase
* the parser state and state modifying functions
* auxiliary functions for parsing
* transformation functions
* parsing
* the definition of the wantModule function

*/


/////////////////////////////////
// Utility functions
//

// '$' is a standard function in Haskell; right associative function application
// example: f $ a $ b $ c $ d = f (a (b (c d)))
// H98, page 93, 99
($) infixr 0
($) a b :== a b

(--) infix 0
(--) a b :== (a,b)

// Replaces sequences of list @arg2 which are identical according to @arg1 whith one element of the sequence.
// example: removeDupBy (==) [1,1,2,2,1] = [1,2,1]
removeDupBy :: (a a -> Bool) !u:[a] -> u:[a]
removeDupBy eq [] = []
removeDupBy eq [h:t] = [h: removeDupBy eq (dropWhile (eq h) t)]

class nothing a :: a

instance nothing String 		where nothing = ""
instance nothing Bool 			where nothing = False
instance nothing (a,b) | nothing a & nothing b where nothing = (nothing,nothing)
instance nothing (a,b,c) | nothing a & nothing b & nothing c where nothing = (nothing,nothing,nothing)
instance nothing (Optional a) 	where nothing = No
instance nothing (Maybe a) 		where nothing = Nothing
instance nothing [a] 			where nothing = []

instance nothing ParsedExpr 	where nothing = PE_Empty
instance nothing ParsedDefinition where nothing = PD_Import []
instance nothing BasicValue 	where nothing = BVS ""
instance nothing Ident 			where nothing = {id_name = "", id_info = nilPtr}
instance nothing TypeVar 		where nothing= { tv_name = nothing, tv_info_ptr = nilPtr }
instance nothing ATypeVar 		where nothing = {atv_attribute = TA_None, atv_variable = nothing}
instance nothing ImportDeclaration where nothing = ID_OldSyntax []
//instance nothing TypeContext 	where nothing = {tc_class= TCClass {glob_object	:: !object,	glob_module	= -1}, tc_types=[], tc_var=nilPtr	}
instance nothing Type 			where nothing= TE
instance nothing HQualifier 	where nothing = HEmptyQualifier
instance nothing LocalDefs 		where nothing = LocalParsedDefs []
instance nothing Priority 		where nothing = NoPrio
//instance nothing OptGuardedAlts where nothing = UnGuardedExpr (make_ewl nothing) //LocalParsedDefs []


class is_ok a :: a-> Bool
//instance is_ok a | nothing a & == a where is_ok a = a<>nothing

instance is_ok String 			where is_ok "" = False ; is_ok _ = True
instance is_ok Bool 			where is_ok a = a
instance is_ok (a,b) | is_ok b 	where is_ok (a,b) = is_ok b
instance is_ok (a,b,c) | is_ok c where is_ok (a,b,c) = is_ok c
instance is_ok (Optional a) 	where is_ok No = False; is_ok _ = True
instance is_ok (Maybe a) 		where is_ok Nothing = False; is_ok _ = True
instance is_ok [a] 				where is_ok [] = False; is_ok _ = True

instance is_ok Haskell_Token	where is_ok (Char_Token "")=False; is_ok _ = True
instance is_ok BasicValue 		where is_ok (BVS "")=False; is_ok _ = True
instance is_ok ParsedExpr 		where is_ok PE_Empty = False; is_ok _ = True
instance is_ok Ident 			where is_ok {id_name = ""} = False; is_ok _ = True
instance is_ok ATypeVar 		where is_ok { atv_variable = x} = is_ok x; is_ok _ = True
instance is_ok TypeVar 			where is_ok { tv_name } = is_ok tv_name
instance is_ok Type 			where is_ok TE= False; is_ok _= True
instance is_ok ParsedDefinition where is_ok (PD_Import [])= False; is_ok _= True
instance is_ok ImportDeclaration where is_ok (ID_OldSyntax [])= False; is_ok _= True
//instance is_ok OptGuardedAlts where /*is_ok TE= False;*/ is_ok _= True  //!!

class dummy a :: a

instance dummy String 			where dummy = "<<dummy>>"
instance dummy (a,b) | dummy a & dummy b where dummy = (dummy,dummy)
instance dummy (Optional a) | dummy a where	dummy = Yes dummy
instance dummy (Maybe a) | dummy a where	dummy = Just dummy
instance dummy [a]  | dummy a 	where dummy = [dummy]

instance dummy ParsedExpr 		where dummy = PE_Basic dummy
instance dummy ParsedDefinition where dummy = PD_Erroneous
instance dummy BasicValue 		where dummy = BVC "<<error>>"
instance dummy Ident 			where dummy = {id_name = "<<error>>", id_info = nilPtr}
instance dummy TypeVar 			where dummy= { tv_name = dummy, tv_info_ptr = nilPtr }
instance dummy ATypeVar 		where dummy = {atv_attribute = TA_None, atv_variable = dummy}
instance dummy Type 			where dummy= TLifted dummy
instance dummy HQualifier 		where dummy = HGuard dummy
instance dummy TypeContext 		where dummy = {tc_class=undef, tc_types=[], tc_var= nilPtr}
instance dummy GuardedExpr 		where dummy = 
											{	alt_nodes	= []
											,	alt_guard	= dummy
											,	alt_expr	= UnGuardedExpr (make_ewl NoPos dummy)
											,	alt_ident	= dummy
											,	alt_position= NoPos
											}


// An instance of file print class (<<<) for printing LineAndColumn positions.
instance <<< LineAndColumn
where
	(<<<) file {lc_line,lc_column} = file <<< lc_line <<< ";" <<< lc_column


class readline a :: *a -> *(!String,!*a)

instance readline File where
	readline f = freadline f

instance readline [String] where
	readline [] = ("",[])
	readline [hd:tl] = (hd,tl)

instance readline String where
	readline s = (s,"")



split ch s
	# end= search_dot (size s- 1)
	= (s%(0,end-1),s%(end+1,size s-1))
where
	search_dot 0 = 0
	search_dot i
		| s.[i]==ch = i
		= search_dot (dec i)



//////////////////////////////////
// The scanner phase
//

/*
The only function exported by this section is:

get_tokens :: *File !Int/*tabsize*/ !Bool/*is_literate*/ -> *Tokens
*/

// type for representing the list of tokens
:: Tokens end
			= No_More_Token !end	// the already scanned file to close

			| Next_Token
				!LineAndColumn			// the beginning position of the token
				!Haskell_Token			// the token
				(Tokens end)

:: Token` :== (!LineAndColumn,!Haskell_Token)

// Token types of language Haskell.
:: Haskell_Token
	// basic values
	= Int_Token		!Int
	| Float_Token	!Int !Int
	| String_Token	!String
	| Char_Token	!String

	// identifiers
	| Var_Id	!String		// varid, tyvar
	| Con_Id	!String		// conid, tycon, tycls, want_modid
	| Var_Sym	!String		// varsym
	| Con_Sym	!String		// consym

	// keywords; if doubt, see the the toString instance
	| Case_Token 		| Class_Token 		| Data_Token 		| Default_Token 	
	| Deriving_Token 	| Do_Token 			| Else_Token 		| If_Token 			
	| Import_Token 		| In_Token 			| Infix_Token 		| Infixl_Token 		
	| Infixr_Token 		| Instance_Token 	| Let_Token 		| Module_Token 		
	| Newtype_Token 	| Of_Token 			| Then_Token 		| Type_Token 		
	| Where_Token 		| WildCard_Token 	

	| DotDot_Token 		| DoubleColon_Token | Equal_Token 		| BackSlash_Token 	
	| Bar_Token 		| LeftArrow_Token 	| Arrow_Token 		| At_Token 			
	| Tilde_Token 		| DoubleArrow_Token 

//	| Colon_Token		// replaced by (Con_Sym ":.")

	| Comma_Token 		| Semicolon_Token 	| Prime_Token 		| SquareOpen_Token 	
	| SquareClose_Token | Open_Token 		| Close_Token 		| CurlyOpen_Token 	
	| CurlyClose_Token 	

	// used in offside handling
	| Newline			 
	| Begin_Group

	// remember
	| Pop
	| Push !Int
	| Mark !Int

	// other auxiliary tokens
	| End_Of_File
	| Error_Token	!String			 

// An instance of toString for Haskell tokens.
instance toString Haskell_Token
where
	toString a = case a of
		Int_Token s		= "integer "+++toString s
		Float_Token a b	= "float "+++toString (toReal a/toReal b)
		String_Token s	= "string "+++s
		Char_Token s	= "char "+++s

		Var_Id s		= "varid "+++s
		Con_Id s		= "conid "+++s
		Var_Sym	s		= "varsym "+++s
		Con_Sym	s		= "consym "+++s

		Case_Token 		= "case"
		Class_Token 	= "class"
		Data_Token 		= "data"
		Default_Token 	= "default"
		Deriving_Token 	= "deriving"
		Do_Token 		= "do"
		Else_Token 		= "else"
		If_Token 		= "if"
		Import_Token 	= "import"
		In_Token 		= "in"
		Infix_Token 	= "infix"
		Infixl_Token 	= "infixl"
		Infixr_Token 	= "infixr"
		Instance_Token 	= "instance"
		Let_Token 		= "let"
		Module_Token 	= "module"
		Newtype_Token 	= "newtype"
		Of_Token 		= "of"
		Then_Token 		= "then"
		Type_Token 		= "type"
		WildCard_Token 	= "_"
		Where_Token 	= "where"
	
		DotDot_Token 	= ".."
		DoubleColon_Token = "::"
		Equal_Token 	= "="
		BackSlash_Token = "\\"
		Bar_Token 		= "|"
		LeftArrow_Token = "<-"
		Arrow_Token 	= "->"
		At_Token 		= "@"
		Tilde_Token 	= "~"
		DoubleArrow_Token = "=>"
	
//		Colon_Token 	= ":"
	
		Comma_Token 	= ","
		Semicolon_Token = ";"
		Prime_Token		= "`"
		SquareOpen_Token = "["
		SquareClose_Token = "]"
		Open_Token 		= "("
		Close_Token 	= ")"
		CurlyOpen_Token = "{"
		CurlyClose_Token = "}"

		Newline			= "newline"
		Begin_Group		= "beginning of a group"
		Pop				= "pop"  //
		Push _			= "push" //
		Mark _			= "mark"
		End_Of_File		= "end of file"
		Error_Token	s	= "scanner error "+++s


// An instance of the (==) operator for Haskell tokens.
instance == Haskell_Token
where
	(==) a b = equal_constructor a b && equal_constructor_args a b
	where
		equal_constructor_args (Int_Token a) (Int_Token b) = a==b
		equal_constructor_args (Float_Token a b) (Float_Token c d) = a*d==b*c
		equal_constructor_args (String_Token a) (String_Token b) = a==b
		equal_constructor_args (Char_Token a) (Char_Token b) = a==b
		equal_constructor_args (Var_Id a) (Var_Id b) = a==b
		equal_constructor_args (Con_Id a) (Con_Id b) = a==b
		equal_constructor_args (Var_Sym a) (Var_Sym b) = a==b
		equal_constructor_args (Con_Sym a) (Con_Sym b) = a==b
		equal_constructor_args (Error_Token a) (Error_Token b) = a==b
		equal_constructor_args _ _ = True
//   or
//import GenEq
//derive gEq Haskell_Token
//instance == Haskell_Token where (==) a b = gEq{|*|} a b

// Used, when the next token is assumed to be an identifier. char_class takes a character (practically the next 
// unanalysed character from the stream) as @arg1, returns the presumed token type of the identifier beginning with this 
// char and a condtion on which kind of characters should the identifer contain. (This will be used for generating
// the token of this identifier.)
char_class :: Char -> (String->Haskell_Token, Char->Bool)
char_class ch
	| ch==':'						= (Con_Sym, is_symbol_char_or_colon)
	| is_symbol_char_or_colon ch	= (Var_Sym, is_symbol_char_or_colon)
	| isUpper ch					= (Con_Id, is_alpha_char_or_apostrophe)
	| otherwise						= (Var_Id, is_alpha_char_or_apostrophe)
where
	is_alpha_char_or_apostrophe :: !Char -> Bool
	is_alpha_char_or_apostrophe c 	= c=='_' || c=='\'' || isAlphanum c

// is_symbol_char_or_colon determines whether character @arg1 can be the part of a symbol or not
// H98, page 5
is_symbol_char_or_colon :: !Char -> Bool
is_symbol_char_or_colon c
	=  c=='!'	|| c=='#'	|| c=='$'	|| c=='%'	|| c=='&'	|| c=='*'	|| c=='+'	|| c=='.'
	|| c=='/'	|| c=='<'	|| c=='='	|| c=='>'	|| c=='?'	|| c=='@'	|| c=='\\'	|| c=='^'
	|| c=='|'	|| c=='-'	|| c=='~'	|| c==':'


// General st of scanning (e.g. the position in the source, the remaining of the source etc.)
:: *ScanState_ a =
			{ input				:: !a
			, line				:: !String	// the actual line
			, row				:: !Int		// the actual row number: 1, 2, ...
			, index				:: !Int		// the actual position in the actual line as a string
			, column			:: !Int		// it equals to the index if there is no tab char in the line

			// static members:
			, tabSize			:: !Int		// it always should be 8
			, literate_comments :: !Bool	// .hs -> False ; .lhs -> True
			}

// get_tokens takes the source code of the file to be analysed as @arg1, the size of tabulators as @arg2 and @arg3 indicating
// that the source code is a literal script or not, and returns the token list generated from the source.
get_tokens :: !Int !Int !Bool !Bool *a -> *Tokens *a | readline a
get_tokens row tabSize literate_comments transform input
	# st	= 	{ input		= input
				, line		= ""
				, row		= row
				, index		= 0
				, column	= 0
				, tabSize	= tabSize
				, literate_comments	= literate_comments
				}
	# tokens= next_line st
	# tokens= if_ transform transform_ids tokens
	= add_newlines_and_begin_groups tokens

//	add_newlines_and_begin_groups` :: Int Bool *(Tokens *a) -> *Tokens *a
transform_ids t=:(Next_Token pos End_Of_File _) = t
transform_ids (Next_Token pos t tokens) = Next_Token pos (f t) (transform_ids tokens)
where
	f (Con_Sym c) = Con_Sym (transform_id '.' c)
	f (Var_Sym c) = Var_Sym (transform_id '.' c)
	f (Con_Id c) = Con_Id (transform_id '_' c)
	f (Var_Id c) = Var_Id (transform_id '_' c)
	f x = x

	transform_id _ "" = ""
	transform_id ch string
		| has_suffix ch string	= string%(0,size string-2)
		= string+++toString ch
	
has_suffix ch str = size str>1 && str.[size str-1]==ch
/*
has_suffix` s str = f (size str-1) (size s-1) where
	f i j
		= j== -1 || i> -1 && s.[j]==str.[i] && f (i-1) (j-1)
*/


// add_newlines_and_begin_groups takes a list of tokens (practically the full token list of the source, since it's called from 
// get_tokens) as @arg1 and inserts Begin_Group tokens in the token list where needed according to the offside rule, and replaces
// a sequence of empty lines and newlines in the source with a single Newline token.
// Real work is done by local function add_newlines_and_begin_groups` which processes the whole token list recursively,
// inserts Begin_Group tokens after each Let_Token, Where_Token, Do_Token and Of_Token and the next token is not CurlyOpenToken
// (in which case the beginning of the group is explicitely marked in the source, so no Begin_Group token is needed.
// The CurlyOpenToken can't be replaced with Begin_Group token, because explicit grouping needs an explicit closing as well, wile
// implicit grouping doesn't.) and inserts Newline tokens each time, when the next token one or more lines under the actual one
// in the source.
add_newlines_and_begin_groups :: *(Tokens *a) -> *Tokens *a
add_newlines_and_begin_groups tokens=:(Next_Token _ first_token _)
	# tokens= add_newlines_and_begin_groups` 0 False tokens
	| first_token==Module_Token || first_token==CurlyOpen_Token		// page 122 !
		= case tokens of (Next_Token _ _ t) = t
	= tokens
where
	add_newlines_and_begin_groups` :: Int Bool *(Tokens *a) -> *Tokens *a
	add_newlines_and_begin_groups` _  was_begin_group t=:(Next_Token pos End_Of_File _)
		|  was_begin_group
			= Next_Token pos Begin_Group t
		= t
	add_newlines_and_begin_groups` previous_row_index was_begin_group (Next_Token pos t tokens)
		# is_begin_group= t==Let_Token || t==Where_Token || t==Do_Token || t==Of_Token
		# tokens= Next_Token pos t (add_newlines_and_begin_groups` pos.lc_line is_begin_group tokens)

		| was_begin_group && t<>CurlyOpen_Token
			= Next_Token pos Begin_Group tokens
		| previous_row_index < pos.lc_line
			= Next_Token pos Newline tokens
		=  tokens


	
// state_to_tokens takes a ScanState_ as @arg1, and uses the first character of the actual line of ScanState to decide, which
// tokenizer function should be used. (cases_by_the_first_char is a hash-like table of tokenizer functions)
state_to_tokens :: !*(ScanState_ *a) -> *Tokens *a | readline a
state_to_tokens st=:{line,index} = cases_by_the_first_char.[toInt line.[index]] st //<<- index

// add_token takes a token as @arg1, an integer as @arg2 and the scan st as @arg3, and inserts the token in the token list 
// stored in the scan st (@arg3). The integer is used to determine where should the next token to be scanned begin. Practically
// used when a string can be replaced with a token. In this case the scanner has to know where this string ends, unless it will
// continue scanning with the next character of the string (e.g. tries to find a token in the middle of an identifier or so). 
add_token :: Haskell_Token Int *(ScanState_ *a) -> *Tokens *a | readline a
add_token token token_end st=:{row,index,column}
	# pos= { lc_line=row, lc_column=column + 1 }
	# st= {st&index=token_end,column= column+token_end-index}
	# tokens= state_to_tokens st
	= Next_Token pos token tokens

//add_token` :: Haskell_Token Int *(ScanState_ *a) -> *Tokens *a | readline a
add_token` pos token token_end st=:{row,index,column}
//	# pos= { lc_line=row, lc_column=column + 1 }
	# st= {st&index=token_end,column= column+token_end-index}
	# tokens= state_to_tokens st
	= Next_Token pos token tokens

cases_by_the_first_char :: {!*(ScanState_ *a)->*Tokens *a} | readline a		// 256 function
cases_by_the_first_char
		=: 	{	( merge_cases
				o removeDupBy (\(x,_) (y,_) = x==y)
				o map remove_first_char
				) item 
			\\ item <-
				( groupBy` (\x= (fst x).[0]) //(\(x,_) (y,_) = x.[0]==y.[0])
				o sortBy (\(x,_) (y,_) = x.[0]<y.[0] || x.[0]==y.[0] && size x>size y)
				) data
			}
where
	remove_first_char (id,fun) = (id%(1,size id-1),fun)

	data = flatten
		[	[	("\xD"		,next_line)
			,	("\xA"		,next_line)
			,	("--"		,one_line_comment)
			,	(" "		,state_to_tokens o skipChar)
			,	("\t"		,state_to_tokens o skipTab)
			,	("{-"		,skip_comment_and state_to_tokens o skipChar o skipChar)
			,	("-}"		,unexpected_end_of_comment)
			,	("\""		,string '\"')
			,	("\'"		,string '\'')
			,	("0o"		,octal_number)
			,	("0O"		,octal_number)
			,	("0x" 		,hexadecimal_number)
			,	("0X" 		,hexadecimal_number)
			,	("()"		,trivial_expression)
			]
		
		, map keyword_
			[	Case_Token		,	Class_Token		,	Data_Token		,	Default_Token  
			,	Deriving_Token  ,	Do_Token  		,	Else_Token  	,	If_Token  		
			,	Import_Token  	,	In_Token  		,	Infix_Token  	,	Infixl_Token  
			,	Infixr_Token 	,	Instance_Token	,	Let_Token  		,	Module_Token  
			,	Newtype_Token	,	Of_Token  		,	Then_Token  	,	Type_Token  
			,	WildCard_Token	,	Where_Token  	,	DotDot_Token	,	DoubleColon_Token  
			,	Equal_Token  	,	BackSlash_Token ,	Bar_Token  		,	LeftArrow_Token  
			,	Arrow_Token		,	At_Token  		,	Tilde_Token		,	DoubleArrow_Token  
//			,	Colon_Token
			]

		, map one_char_keyword
			[	Comma_Token			,	Semicolon_Token		,	Prime_Token			,	SquareOpen_Token
			,	SquareClose_Token	,	Open_Token			,	Close_Token			,	CurlyOpen_Token
			,	CurlyClose_Token	]
	
		, map ident 
			['ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_!#$%&*+./<=>?@\\^|-~:']

		, [ (toString digit, number) \\ digit <- ['0123456789'] ]
		, [ (toString (toChar i), unknown_char) \\ i<- [0..255] ]
		]

	merge_cases :: ![(!String,!(*(ScanState_ *a)->*Tokens *a))] -> *(ScanState_ *a)->*Tokens *a | readline a
	merge_cases [("",fun)] = fun
	merge_cases [(str,fun1),("",fun2)] = fun
	where
		fun st=:{line,index}
			| match_string str line (index+1) = fun1 st
			= fun2 st
	merge_cases [(str,fun1):tl] = fun
	where
		other_cases = merge_cases tl

		fun st=:{line,index}
			| match_string str line (index+1) = fun1 st
			= other_cases st

	match_string str line col = f col 0 where
		f col c = c==size str || str.[c]==line.[col] && f (inc col) (inc c)

unknown_char st=:{line,index}
	= add_token (Error_Token $ "Unknow char: "+++toPrint line.[index]) (index+1) st

toPrint '\'' = "\\\'"
toPrint '\"' = "\\\""
toPrint '\\' = "\\\\"
toPrint ch
	| not (isPrint ch) = toString ['\\' : toOct (toInt ch)]
	= toString ch
where
	toOct 0 = ['0']
	toOct i = f [] i where
		f acc 0 = acc
		f acc x = f [toChar (x rem 8 + toInt '0'):acc] (x>>3)

skipBy cond index line = skip index where
	skip index
		| cond line.[index]
			= skip (index + 1)
		= index

one_line_comment st=:{line,index}
	# end= skipBy ((==)'-') (index+2) line
	# end`= skipBy is_symbol_char_or_colon end line
	| end==end`
		= next_line st
	# str= line%(index,end`-1)
	= add_token (Var_Sym str) end st

one_char_keyword try_token = (toString try_token, f) where f st=:{index}= add_token try_token (index+1) st

trivial_expression st=:{index}= add_token (Con_Id "Trivial") (index+2) st

keyword_ try_token = (kw, f) where

	kw = toString try_token
	keyword_size = size kw
	(ident_type,cond) = char_class kw.[0]

	f st=:{line,index}
		# end = index + keyword_size
		| not (cond line.[end])
			= add_token try_token end st
		# end= skipBy cond end line
		= add_token (ident_type $ line%(index,end-1)) end st

ident ch = (toString ch, f) where
	(ident_type, cond)= char_class ch

	f st=:{line,index}
		# end= skipBy cond (inc index) line
		| isUpper ch && line.[end]=='.' && size line>end+1 && isAlpha line.[end+1]
			= add_token (Error_Token "Qualified name") (end+1) st
		# str= line%(index,end-1)
		= add_token (ident_type str) end st

skipChar st :== {st&index=st.index+1, column=st.column+1}
skipChar` n st :== {st&index=st.index+n, column=st.column+n}
skipTab st :== {st& column= st.tabSize*(st.column/st.tabSize+1), index=st.index+1}

skip_comment_and :: !(*(ScanState_ *a) -> *Tokens *a) !*(ScanState_ *a) -> *Tokens *a | readline a
skip_comment_and continue st=:{line,index}
	# c1 = line.[index]
	# c2 = line.[index+1]
	| c1=='\xD'	|| c1=='\xA'	= read_line_and (skip_comment_and continue) st
	| c1=='-' && c2=='}'		= continue (skipChar (skipChar st))
	| c1=='{' && c2=='-'		= skip_comment_and (skip_comment_and continue) (skipChar (skipChar st))
	| c1=='\t'					= skip_comment_and continue (skipTab st)
	| otherwise					= skip_comment_and continue (skipChar st)

octal_number st=:{index,line}
	| isOctDigit line.[index+2]
		# end= skipBy isOctDigit (index+3) line
		= add_token (Int_Token $ toInt $ "0"+++line%(index+2,end-1)) end st
	= add_token (Error_Token "Octal digit expected") (index+2) st

hexadecimal_number st=:{index,line}
	| isHexDigit line.[index+2]
		# end= skipBy isHexDigit (index+3) line
		= add_token (Int_Token $ toInt $ line%(index,end-1)) end st
	= add_token (Error_Token "Hexadecimal digit expected") (index+2) st

number st=:{index,line}
	# end= skipBy isDigit (index+1) line
	# n= toInt $ line%(index,end-1)
	| line.[end]=='.' && isDigit line.[end+1]
		# end`= skipBy isDigit (end+1) line
		# m= toInt $ line%(end+1,end`-1)
		# l= end`-end-1
		# (k,end`,ok)= ff with
			ff
				| line.[end`]=='E'
					# end`= end`+1
					# ch= line.[end`]
					| ch=='-' || ch=='+'
						# end= end`+1
						| isDigit line.[end]
							# end= skipBy isDigit (end+1) line
							# k= toInt $ line%(end`,end-1)
							= (k,end,True)
						= (0,end,False)
					# end= end`
					| isDigit line.[end]
						# end= skipBy isDigit (end+1) line
						# k= toInt $ line%(end`,end-1)
						= (k,end,True)
					= (0,end,False)
				= (0,end`,True)
		| not ok
			= add_token (Error_Token "Digit expected") end` st
		= add_token (make_float` n m l k) end` st
	= add_token (Int_Token n) end st
where
	float_end x e2 end st
		| isDigit line.[end]
			# end= skipBy isDigit (end+1) line
			# k= toInt $ line%(e2,end-1)
			= add_token (make_float` x k) end st
		= add_token (Error_Token "Digit expected") end st

make_float` n m l k
	# n= n*10^l+m
	# d= l-k
	| d<0	= Float_Token (n*10^(~d)) 1
	= Float_Token n (10^d)

unexpected_end_of_comment st=:{index}
	= add_token (Error_Token "Unexpected end of comment.") (index+2) st

// some additional work needed...
string end_ch st=:{index,row,column}
	= cyc [] (skipChar st)
where
	end_str = toString end_ch

	pos= { lc_line=row, lc_column=column + 1 }

	cyc :: [Char] *(ScanState_ *a) -> *Tokens *a | readline a
	cyc acc st=:{line,index} = case line.[index] of
		'\\'
			# ch= line.[index+1]
			# st= skipChar st
			| ch==' ' || ch=='\t' || ch=='\n'	= skip_gap st with
				skip_gap :: *(ScanState_ *a) -> *Tokens *a | readline a
				skip_gap st=:{line,index} = case line.[index] of
					'\n'	= read_line_and skip_gap st
					'\t'	= skip_gap (skipTab st)
					' '		= skip_gap (skipChar st)
					'\\'	= cyc acc (skipChar st)
					ch		= add_token (Error_Token $ "Something inside gap." +++ toString (toInt ch)) (size line-1) st
			| isDigit ch
				# (d,st)= digits isDigit st
				= cyc [toChar (c_val 10 d):acc] st
			# st= skipChar st
			= case ch of
				'&'	= cyc ['&':acc] st
				'x' # (d,st)= digits isHexDigit st
					= cyc [toChar (c_val 16 d):acc] st
				'o' # (d,st)= digits isOctDigit st
					= cyc [toChar (c_val 8 d):acc] st
					
				_	= case [(val,size str) \\ (str,val)<- ascii | str==line%(index+1,index+size str)] of
						[] 		= add_token (Error_Token ("Invalid char: " +++ toPrint ch)) (size line-1) st
						[(val,s):_] 	= cyc [val:acc] (skipChar` (s-1) st)

		ch | ch==end_ch	= add_token` pos (if (end_ch=='\'') Char_Token String_Token (foldl (+++) "" [end_str: reverse [end_str:map toPrint acc]])) (index+1) st
		'\n'	= add_token (Error_Token "End of line inside string.") (size line-1) st
		'\t'	= cyc ['\t':acc] (skipTab st)
		ch		= cyc [ch:acc] (skipChar st)

	ascii = zip2
			   ["NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL",
			    "BS",  "HT",  "LF",  "VT",  "FF",  "CR",  "SO",  "SI",
			    "DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN", "ETB",
			    "CAN", "EM",  "SUB", "ESC", "FS",  "GS",  "RS",  "US",
			    "SP"]  ['\0'..]
	   ++ [("DEL",'\b'),("a",'\a'),("b",'\b'),("f",'\f'),("n",'\n'),("r",'\r')
	      ,("t",'\t'),("v",'\v'),("\\",'\\'),("\"",'\"'),("\'",'\'')]

	c_val x [] = 0
	c_val x [a:b] = a + x * c_val x b
	
	digits p st= f [] st where
		f acc st=:{line,index}
			| p line.[index] = f [g line.[index]:acc] (skipChar st)
			= (acc,st)
		g x | isDigit x = toInt x - toInt '0'
		    | isUpper x = toInt x - toInt 'A' + 10
		                = toInt x - toInt 'a' + 10

next_line :: *(ScanState_ *a) -> *Tokens *a | readline a
next_line x = read_line_and state_to_tokens x

read_line_and continue st=:{input,row,literate_comments}
	#! (line,input)= readline input
	| line==""
		= Next_Token {lc_line=row,lc_column= 1} End_Of_File (No_More_Token input)

	# end = f (size line-1) with
		f -1 = -1
		f i	= case line.[i] of
			' '   = f (i-1)
			'\xD' = f (i-1)
			'\xA' = f (i-1)
			'\t'  = f (i-1)
			_     = i
			
	#! line = line%(0,end) +++ "\n"
	#! line= case literate_comments of
		True
			| line.[0] == '>'
				= " "+++line%(1,size line-1)
			= "\n"
		_	= line

	= continue {st& line= line, row=inc st.row, input=input, index=0, column=0}
	


////////////////////////////////////////////////////////////////////
// The parser st and st modifying functions
//

:: *State a =
			{ input_name 		:: !String
			, iclmodule			:: !Bool
			
			, tokens			:: *Tokens a
			, layout_contexts_stack	:: !*[Int]
			, last_tokens		:: !*[!Token`!]

			, error_file		:: !*File
			, skipping			:: !Bool
			, okey				:: !Bool

			, hash_table			:: !*HashTable
			, id_count				:: !Int
			}

parse_tokens_with :: (*(State *b)->(!c,!*(State *b))) *b *(State .a) -> *(!c,!*(State .a)) | readline b
parse_tokens_with p input st=:{tokens=t=:Next_Token {lc_line} _ _,layout_contexts_stack,last_tokens}
	# st= {st& tokens= get_tokens lc_line 8 False False input, layout_contexts_stack= [0], last_tokens= [!!]}
	# (x,st)= p st
	= x -- {st& tokens=t, layout_contexts_stack=layout_contexts_stack, last_tokens=last_tokens}

open_state input_file file_name iclmodule tabsize error_file hash_table
	# lhs= False //!!! size file_name >7 && file_name.[size file_name-7]=='l'
	=	{ tokens			= get_tokens 0 tabsize lhs True input_file
		, layout_contexts_stack	= [0]
		, last_tokens		= [!!]

		, error_file		= error_file
		, okey				= True
		, skipping			= False

		, input_name		= file_name
		, iclmodule			= iclmodule

		, hash_table		= hash_table
		, id_count			= 0
		}

close_state = 
	(iff` End_Of_File 
		id
		(expected "end of file")
	) .> close
where
	close {okey, tokens, error_file, hash_table}
		= (okey, get_file tokens, error_file, hash_table)

	get_file t = case t of
		No_More_Token f = f
		Next_Token _ _ t = get_file t


get_new_id` pre st=:{id_count}
	# id = pre+++"_"+++toString id_count
	= (id,{st& id_count= id_count+1})

writeMessage f st	:== {st&error_file= f st.error_file}
after_error :== snd o try_token_by id

is_icl :: *(State .a) -> *(!Bool,!*(State .a))
is_icl st=:{iclmodule} = (iclmodule,st)

only_def :: *(State .a) -> *(State .a)
//only_def st=:{hash_table}= {st& hash_table= set_hte_mark 0 hash_table}
only_def st=:{hash_table}= st// !!!

imp_or_def :: *(State .a) -> *(State .a)
imp_or_def st=:{hash_table,iclmodule}= {st& hash_table= set_hte_mark (if iclmodule 1 0) hash_table}

into_hashTable :: !String !IdentClass !*(State .a) -> *(!Ident,!*(State .a))
into_hashTable string ident_class st=:{hash_table}
	#! ({boxed_ident=ident}, hash_table) = putIdentInHashTable string ident_class hash_table
	= (ident,{st&hash_table=hash_table})

getPos			:: !*(State .a) -> *(!Position,!*(State .a)) 
getPos st=:{input_name,tokens= Next_Token {lc_line} _ _} = (LinePos input_name lc_line, st)

getLineAndColumn			:: !*(State .a) -> *(!LineAndColumn,!*(State .a)) 
getLineAndColumn st=:{tokens= Next_Token pos _ _} = (pos, st)

protect p st
	# (id,st)= remember st
	# (x,st)= p st
	| is_ok x
		= (x, st)
	= (x, tokens_back id st)
where
//	remember :: *(State .a) -> (!Int,!*(State .a))
	remember st=:{id_count, last_tokens, tokens= Next_Token pos _ _}
		= (id_count, {st& id_count= id_count+1, last_tokens=[!(pos, Mark id_count):last_tokens!]})
	
//	tokens_back :: Int *(State .a) -> *(State .a)
	tokens_back id st= f st where
		f st=:{layout_contexts_stack=l, tokens, last_tokens=[!(pos, token):last_tokens!]}
			= case token of
				Pop	
							= f {st& layout_contexts_stack=tl l, last_tokens=last_tokens}
				Push x
							= f {st& layout_contexts_stack=[x:l], last_tokens=last_tokens}
				Mark x
							# st= {st& last_tokens=last_tokens}
							| id==x = st
							= f st
				token	
							= f {st& tokens=Next_Token pos token tokens, last_tokens=last_tokens}
		f st 	= error` "tokens_back" st

//get_a_token st=:{id_count} | id_count>1000000 = abort "maybe cycle"
//											= get_a_token` {st& id_count=id_count+1}


//try_token_by :: !(Haskell_Token->x) !*(State .a) -> *(!x,!*(State .a)) | is x	//CurlyClose
try_token_by f st
	# (token,g,st)= get_a_token st
	# a= f token
	| is_ok a
		= (a, g st)
	= (a, st)
where
	get_a_token :: *(State .a) -> *(Haskell_Token,*(State .a)->*(State .a),*(State .a))
	get_a_token st=:{layout_contexts_stack=[hd:tl], tokens= Next_Token {lc_column} token _}
		= case token of
			Newline			| hd==lc_column		= (Semicolon_Token, skip_tok, st)
							| hd>lc_column		= (CurlyClose_Token, dec_stack, st)
												= get_a_token (skip_tok st)	///
			Begin_Group		| hd<lc_column		= (CurlyOpen_Token, add_to_stack lc_column o skip_tok, st)
												= get_a_token (error` "Note 3" (skip_tok st))			// Note 3
			End_Of_File		| hd<>0				= (CurlyClose_Token, dec_stack, st)
			 				| tl==[]			= (End_Of_File, id, st)
			 									= get_a_token (error` "Note 5" (skip_tok st))			// Note 5, p123
			CurlyClose_Token| tl==[]			= get_a_token (error` "missing curly open" (skip_tok st))
							| hd==0				= (CurlyClose_Token, dec_stack o skip_tok, st)
			 									= get_a_token (error` "Note 2" (skip_tok st))			// Note 2
			Error_Token msg=:"Qualified name"	= get_a_token (warning msg (skip_tok st))
			Error_Token msg						= get_a_token (error` msg (skip_tok st))
			CurlyOpen_Token						= (CurlyOpen_Token, add_to_stack 0 o skip_tok, st)
			token								= (token, skip_tok, st)
	

skip_tok st=:{tokens=Next_Token _ End_Of_File _}	= error` "no more token" st
skip_tok st=:{tokens=Next_Token pos token tokens, last_tokens} = {st& tokens=tokens, last_tokens=[!(pos,token):last_tokens!]}

force_curlyClose :== dec_stack

add_to_stack x st=:{layout_contexts_stack=s, last_tokens, tokens= Next_Token pos _ _} = {st& layout_contexts_stack=[x:s], last_tokens=[!(pos,Pop):last_tokens!]}

dec_stack :: *(State .a) -> *(State .a)
dec_stack st=:{layout_contexts_stack=[_]} = error` "missing curly open" st
dec_stack st=:{layout_contexts_stack=[hd:tl], last_tokens, tokens= Next_Token pos _ _}
	= {st& layout_contexts_stack= tl, last_tokens=[!(pos,Push hd):last_tokens!]}

expected		:: !x !*(State .a) -> *(State .a)	| toString x
expected message st=:{input_name}
	# (token,st)= try_token_by id st
	= error (toString message +++ " expected instead of " +++ toString token) st

error				:: !String !*(State .a) -> *(State .a)
error message st= after_error (error` message st)

error`				:: !String !*(State .a) -> *(State .a)
error` message st
	# (position, st)= getLineAndColumn st
	= error`` position message st

error`` _ message st=:{skipping=True} = st
error`` position message st=:{input_name}	= writeMessage h {st& okey=False, skipping=True}
where
	h x= x	<<< "Parse error [" <<< input_name <<< ',' <<< position <<< "]: " <<< message <<< "\n"

warning`` message st=:{iclmodule=False} = st
warning`` message st = warning message st

warning :: !String !*(State .a) -> *(State .a)
warning message st
	# (position, st)= getLineAndColumn st
	= warning` position message st

warning` _ message st=:{skipping=True} = st
warning` position message st=:{input_name}	= writeMessage h st
where
	h x= x	<<< "Parse warning [" <<< input_name <<< ',' <<< position <<< "]: " <<< message <<< "\n"

clearError st	:== {st& skipping=False}

want_end_of_def st
	= try_token Semicolon_Token st
/*	# (yes,st)= try_token Semicolon_Token st
	| yes	= (True, clearError st)
	# (yes,st)= check_token CurlyClose_Token st
	| yes	= (False, clearError st)
	# st= expected "end of definition" st
	# (yes,st)= check_token End_Of_File st
	| yes	= (False,st)
	# st= skip_current_token st
	= want_end_of_def st
*/	


want_token :: !Haskell_Token !*(State .a) -> *(State .a)	 
want_token CurlyClose_Token st
	# (yes,st)= try_token CurlyClose_Token st
	| yes	= st
	= force_curlyClose st
want_token t st
	# (yes,st)= try_token t st
	| yes	= st
	= expected (toString t) st

try_token t = try_token_by ((==)t)

	
////////////////////////////////////////////////////////
// Auxiliary functions for parsing
//

(.>) infixr 5
(.>) a b = \st= b (a st)

(..>) infixr 5
(..>) a b :== a .> uncurry b

(?>) infixr 5
(?>) a b :== a .> if_ok b

(?.>) infixr 5
(?.>) a b :== a .> if_ok (uncurry b)

($.) infixr 5
($.) a b :== try_token a ?> snd .> b //

(.$) infixr 5
(.$) a b :== a .> app_st (want_token b)

if_ok b (a, st) | is_ok a 	= b (a, st)
if_ok _ (_, st)				= nothing -- st

app_val f (a, st) 	:== (f a, st)
app_st f (a, st) 	:== (a, f st)

add_val b (a, st) = case b st of b -- st = (a,b) -- st

add_val` b (a, st) = case b st of
		b -- st	| is_ok b	= a -- st
		b -- st 			= nothing -- st


(!.) infix 6
(!.) a b :== a .> to_want b

to_want _ (x, st) | is_ok x = x -- st
to_want m (x, st) = dummy -- expected m st

iff p q r st
	# (x,st) = p st
	| is_ok x
		= q x st
	= r st

iff` p q r st
	# (x,st) = try_token p st
	| is_ok x
		= q st
	= r st


(<|>) infixr 3 //:: (.a -> (!Optional b,!.a)) (.a -> (!Optional b,!.a)) -> (.a -> (!Optional b,!.a))
(<|>) a b = \st = case a st of
		a -- st	| is_ok a 	= a -- st
		a -- st				= b st

add a b = a -- b

cons_ hd (tl,st) = [hd:tl] -- st
//cycle p (hd,st) = case p st of (tl,st) = [hd:tl] -- st

//zeroOrMore  :: !(*(State .a) -> *(!Optional x,*(State .a))) !*(State .a) -> *(![x],!*(State .a))	 
zeroOrMore pp = p where
	p	= pp ?.> \hd = p .> cons_ hd

oneOrMore p msg = p !. msg ..> \hd= zeroOrMore p .> cons_ hd

listWith pp comma = p where
	p = pp ..> \hd= 
			comma $. p .> cons_ hd
		<|>	add [hd]

commaList pp = p where
	p = pp ..> \hd= 
			Comma_Token $. p .> cons_ hd
		<|>	add [hd]

commaList` pp = p where
	p = pp ?.> \hd= 
			Comma_Token $. p ?> cons_ hd
		<|>	add [hd]

commaList_or_nothing` p msg = 
		p ?.> (\hd= 
			Comma_Token $. commaList (p !. msg) ?> cons_ hd
		<|>	add [hd]
		)
	<|> add [] //!!!

groupWith pp = want_token CurlyOpen_Token .> p .$ CurlyClose_Token where
	p = pp ..> \x= 
			Semicolon_Token $. p ..> (\y= add [x:y])
		<|>	add [x]


if_ bool f st :== if bool (f st) st



/////////////////////////////////////////////////
// Transformation functions
//


make_application name args st
	# (f, st)= make_ident name st
	= (PE_List [f: args], st)

make_exp [x] = x
make_exp l = PE_List l

make_list l = add (foldr make_node make_nil l)
make_ident a = into_ht` IC_Expression a .> app_val PE_Ident


//remove_TA_None {at_type} = at_type
add_TA_None t :== {at_attribute= TA_None, at_type= t}

makeParsedImport pos mod_ident import_symbols	:== { import_module = mod_ident, import_symbols = import_symbols, import_file_position = pos }
makeListTypeSymbol t 		:== TA (MakeNewTypeSymbIdent predefined_idents.[PD_ListType] (length t)) t
NoLocalParsedDefs			:== LocalParsedDefs []

make_btype :: ![Type] !*(State .a) -> *(!Type,!*(State .a))	 
make_btype [type] st = (type,st)
make_btype [type: types] st	= case type of
	TA sym []	= (TA { sym & type_arity = length types } (map add_TA_None types), st)
	TV tv		= (CV tv :@: (map add_TA_None types), st)
	TArrow		= case types of
		[type1, type2]	= (add_TA_None type1 --> add_TA_None type2, st)
		[type1]			= (TArrow1 (add_TA_None type1), st)
		_ = cc
	TArrow1 type1	= case types of
		[type2]	= (type1 --> add_TA_None type2, st)
		_ = cc
	_ = cc
where
	cc = (type, error "Type;ordinary type variable" st)



lift_function l=:[PD_Function pos name is_infix [] rhs fk:_] st = (l,st)
lift_function l=:[typeSpec=:PD_TypeSpec _ _ _ _ _,PD_Function pos name is_infix [] rhs fk:_] st = (l,st)
lift_function [typeSpec=:PD_TypeSpec _ _ _ (Yes _) _:l=:[PD_Function pos name False args rhs fk:_]] st
	# name`= toString name
	# name`= name`+++if (size name`>0 && is_symbol_char_or_colon name`.[0]) "*" "8"
	# (name`,st)= into_ht` IC_Expression name` st
//	# name`= name
	# rhs` =
		{ rhs_alts = UnGuardedExpr (make_ewl NoPos (PE_Ident name`))
		, rhs_locals = LocalParsedDefs (map (\(PD_Function pos _ _ args rhs _)= PD_Function pos name` False args rhs (FK_Fun)) l)
		}

	# function= PD_Function pos name False [] rhs` fk
	= ([typeSpec,function], st)
lift_function l=:[PD_Function pos name False args rhs fk:_] st
	# name`= toString name
	# name`= name`+++if (size name`>0 && is_symbol_char_or_colon name`.[0]) "*" "8"
	# (name`,st)= into_ht` IC_Expression name` st
//	# name`= name
	# rhs` =
		{ rhs_alts = UnGuardedExpr (make_ewl NoPos (PE_Ident name`))
		, rhs_locals = LocalParsedDefs (map (\(PD_Function pos _ _ args rhs _)= PD_Function pos name` False args rhs (FK_Fun)) l)
		}

	# function= PD_Function pos name False [] rhs` fk
	= ([function], st)
lift_function l st = (l,st) //abort "lift_function"

pd_name (PD_TypeSpec _ n _ _ _) = toString n
pd_name (PD_Function _ n _ _ _ _) = toString n
pd_name _ = ""

groupBy` f l = groupBy (\x y=f x==f y) l
sortBy` f l = sortBy (\x y=f x==f y) l

is_ts (PD_TypeSpec _ n _ _ _) = True
is_ts _ = False
is_class (PD_Class _ _) = True
is_class _ = False
is_ins (PD_Instance _) = True
is_ins _ = False

//bind_types_and_bodies :: !*[ParsedDefinition] !*(State .a) -> *(!*[ParsedDefinition],!*(State .a))
bind_types_and_bodies defs st
	# (ty,defs)= partition is_ts defs
	# defs= groupBy` pd_name (foldlSt (insert p) ty defs) with
		p a b=pd_name a==pd_name b

	# (defs,st)= maplSt lift_function defs st
	# defs= flatten defs
	= (defs, st)

ch_pos pos (PD_Function a b c d e f)= PD_Function pos b c d e f

//context_name {tc_class=TCClass {glob_object={ds_ident}}} = toString ds_ident

get_new_id st = get_new_id` "a" st
get_new_ids l st
	# (id,st) = get_new_id` "" st
	= (take l [toString c+++id \\ c<-['a'..]], st)


isSymOp (PE_Ident x)
	# x= toString x
	= has_suffix '?' x || size x>1 && x.[0]==':'
isSymOp _ = False
/*
is_varid (PE_Ident .>)
	# .>= toString .>
	= size .>>0 && isLower .>.[0] //!!!
is_varid _ = False
*/
is_con (PE_Ident x)
	# x= toString x
	= size x>0 && isUpper x.[0]
is_con _ = False

is_appl s
	# s= toString s
	= has_suffix '_' s && size s>0 && not (isUpper s.[0])

is_conop (PE_Ident x)
	# x= toString x
	= size x>0 && x.[0]==':'
is_conop _ = False

isOp (PE_Ident x)
	# x= toString x
	= has_suffix '?' x || has_suffix '9' x || size x>1 && x.[0]==':'
isOp _ = False

make_tuple l= case l of
				[PE_Ident _]	= PE_List l
				[x]				= x
				l				= PE_Tuple l

add_instance_defs defs st
	# (icl,st)= is_icl st
	# (classes,defs)= partition is_class defs
	# (instances,defs)= partition is_ins defs
	# (classes,(instances,st))= maplSt f classes (instances,st) with

		f (PD_Class def=:{class_name, class_arity, class_context, class_args=[arg]} members) (instances,st)
//			# members= groupBy` pd_name members
			# (types,macros)= partition is_ts members
			# (ins,instances) = partition (\(PD_Instance {pi_class})= pi_class==class_name) instances
			| not icl
				= [PD_Class def types: ins] -- (instances,st)
			# (ins,st)= maplSt g ins st with
				g (PD_Instance def=:{pi_members,pi_pos,pi_types=[type]}) st
					# names= map pd_name pi_members
					# pi_members= [ch_pos pi_pos m \\ m<-macros | not (isMember (pd_name m) names)] ++ pi_members
//					# (pi_members,st)= maplSt lift_function pi_members st
//					# (pi_members,st)= bind_types_and_bodies pi_members st //(types++pi_members) st
//					# pi_members= filter (not o is_ts) pi_members
					= (PD_Instance {def& pi_members=pi_members}, st)
				g x st = abort "add_i"
			= [PD_Class def types: ins] -- (instances,st)
		f _ _ = abort "add_i2"

	= flatten classes ++ instances ++ defs -- st
//	= defs //classes ++ instances ++ defs


buildRhs rhs =
	{ rhs_alts = UnGuardedExpr (make_ewl NoPos rhs)
	, rhs_locals = LocalParsedDefs []
	}

make_nil =: PE_Ident predefined_idents.[PD_NilSymbol]
make_node hd tl= PE_List [PE_Ident predefined_idents.[PD_ConsSymbol],hd,tl]

transform_do :: [HQualifier] *(State .a) -> *(!ParsedExpr,!*(State .a))
transform_do [HGuard e] st = (e, st)
transform_do [HEmptyQualifier:tl] st = transform_do tl st
transform_do [HGuard x:tl] st
	# (y, st)= transform_do tl st
	= make_application ">>." [x, y] st
transform_do [HGenerator pos pos2 patt list: tl] st
	# (rhs, st)= transform_do tl st
	# (le,gu,patt,st)= tr_patt patt st
	| isNothing gu && always_match patt
		# (lam_ident,st)= internalIdent` "\\" pos2 st
		= make_application ">>=." [list, PE_Lambda lam_ident [patt] rhs pos] st

	# (ok_ident,st)= internalIdent` "g_f" pos2 st
	# (rhs1,st)= buildRhs` le gu rhs st
	# (rhs2,st)= parse_tokens_with (try_rhs` (ok_ident,Yes [PE_WildCard])) "=fail\"error!\"" st
	# defs= LocalParsedDefs
		[	PD_Function pos ok_ident False [patt] rhs1 (FK_Fun)
		:	rhs2 //PD_Function pos ok_ident False [PE_WildCard] rhs2 (FK_Fun)
		]
	# (exp, st)= make_application ">>=." [list, PE_Ident ok_ident] st
	= (PE_Let False defs exp, st)
where
	always_match :: ParsedExpr -> Bool
	always_match (PE_Ident _) = True
	always_match (PE_Tuple l) = all always_match l
	//...
	always_match _ = False
	
transform_do [HLocal defs: tl] st
	# (exp, st)= transform_do tl st
	= (PE_Let False defs exp, st)
transform_do _ st = (PE_Empty, error "#tr_do" st)

transform_qualifier exp [] st = (make_node exp make_nil, st)
transform_qualifier exp [HEmptyQualifier:tl] st = transform_qualifier exp tl st
transform_qualifier exp [HGuard b:tl] st
	# (if_ident,st)= internalIdent "_if" st
	# (then_exp, st)= transform_qualifier exp tl st
	= (PE_If if_ident b then_exp make_nil, st)
transform_qualifier exp [HGenerator pos pos2 patt list: tl] st
	# (le,gu,patt,st)= tr_patt patt st
	| isNothing gu && always_match patt
		# (lam_ident,st)= internalIdent` "\\" pos2 st
		# (rhs, st)= transform_qualifier exp tl st
		# (exp, st)= make_application "concatMap" [PE_Lambda lam_ident [patt] rhs pos, list] st
		= (exp, st)

	# (ok_ident,st)= internalIdent` "g_f" pos2 st
	# (rhs, st)= transform_qualifier exp tl st
	# (rhs1,st)= buildRhs` le gu rhs st
	# defs= LocalParsedDefs
		[	PD_Function pos ok_ident False [patt] rhs1 (FK_Fun)
		,	PD_Function pos ok_ident False [PE_WildCard] (buildRhs make_nil) (FK_Fun)
		]
	# (exp, st)= make_application "concatMap" [PE_Ident ok_ident, list] st
	= (PE_Let False defs exp, st)
where
	always_match :: ParsedExpr -> Bool
	always_match (PE_Ident _) = True
	always_match (PE_Tuple l) = all always_match l
	//...
	always_match _ = False
	
transform_qualifier exp [HLocal defs: tl] st
	# (exp, st)= transform_qualifier exp tl st
	= (PE_Let False defs exp, st)

::	HQualifier
	=	HGenerator Position LineAndColumn ParsedExpr /*(Maybe ParsedExpr)*/ ParsedExpr
	|	HLocal LocalDefs
	|	HGuard ParsedExpr
	|	HEmptyQualifier

make_fixitydef pos prio args name st
	| size name>0 && name.[0]==':'
		= ([],st)
	# name`= case has_suffix '.' name of
		True	= name+++"?"
		_		= name+++"9"
	# (name`,st) = into_hashTable name` IC_Expression st
	# (expr,st)= make_application name args st
	=  ([ PD_TypeSpec pos name` prio No SP_None
		, PD_Function pos name` True args (buildRhs expr) FK_Macro //FK_Fun
		], st)

FK_Fun :== FK_Function cNameNotLocationDependent

//buildNodeDef :: ParsedExpr ParsedExpr -> ParsedDefinition
//buildNodeDef lhsExpr rhsExpr =	PD_NodeDef NoPos lhsExpr (buildRhs rhsExpr)

internalIdent :: !String !*(State .a) -> (!Ident, !*(State .a))	 
internalIdent prefix st
	# (lc,st)	= getLineAndColumn st
	= internalIdent` prefix lc st

internalIdent` :: !String !LineAndColumn !*(State .a) -> (!Ident, !*(State .a))	 
internalIdent` prefix {lc_line,lc_column} st
	# ident	= prefix +++ ";" +++ toString lc_line +++ ";" +++ toString lc_column
	= into_hashTable ident IC_Expression st


make_deriv type`=:{td_pos=pos, td_name=type_name, td_arity=type_arity, td_rhs = ConsList list} names st
	# list= [(toString pc_cons_name,pc_cons_arity,f pc_cons_prio)\\ {pc_cons_name,pc_cons_arity,pc_cons_prio}<-list] with
		f NoPrio = False
		f _ = True
	# tn= toString type_name
	# type_symb	= MakeNewTypeSymbIdent type_name 0
	# type		= TA type_symb []

	# (decls,st)= maplSt f names st with
//		f (name,pos`) st | isMember name ["Read_","Show_","Ix_"] = ([], warning` pos` (name+++" could't be derived") st)
		f (name,pos`) st

//			# name`= transform_id name
			# (id,st)= get_new_id st
			# args= take type_arity [id+++toString ch \\ ch<- ['a'..]]
			# prg= case name of
				"Eq_"
					# list`= map f list with
						f (n,0,_) = n+++" ==. "+++n+++" =True_;"
						f (n,2,True)= "a "+++n+++" b==.c "+++n+++" d=a==.c&&b==.d;"
						f (n,i,_)= concat ([n]++[" a"+++toString j\\j<-[1..i]]++[" ==. ",n]++[" b"+++toString j\\j<-[1..i]]++["="]++intersperse"&&"["a"+++toString j+++"==.b"+++toString j\\j<-[1..i]]++[";"])
					= list` ++ ["_==._=False_;","x/=.y=not_(x==.y)"]
				"Ord_"
					# list`= map f list with
						f (n,2,True)= "a "+++n+++" b<.c "+++n+++" d=a<.c||a==.c&&b<.d;"
						f (n,i,_)
							# xx= foldr g ["False_"] [("a"+++toString j,"b"+++toString j)\\j<-[1..i]] with
								g (a,b) l = [a,"<.",b,"||",a,"==.",b,"&&(":l]++[")"]
							= concat ([n]++[" a"+++toString j\\j<-[1..i]]++[" <. ",n]++[" b"+++toString j\\j<-[1..i]]++["="]++xx++[";"])
					# last= ["a<.b=f a<.f b where{"]++ll++["};"] with
						ll = intersperse ";" [f i it\\i<-[0..] & it<-list]
						f i (n,2,True)= "f(_ "+++n+++" _)="+++toString i+++"::Int_"
						f i (n,i`,_)="f("+++n+++concat (repeatn i` " _")+++")="+++toString i+++"::Int_"

					= list` ++ last ++ 
						[	"compare_ x y|x==.y=EQ_|x<.y=LT_|otherwise_=GT_;"
						,	"x<=.y=compare_ x y/=.GT_;"
						,	"x>=.y=compare_ x y/=.LT_;"
    					,	"x>.y=compare_ x y==.GT_;"
						,	"max_ x y|x<=.y=y|otherwise_=x;"
						,	"min_ x y|x<=.y=x|otherwise_=y"]

				"Bounded_"
					= ["minBound_=",f " minBound_ " (hd list),";maxBound_=",f " maxBound_ " (last list)] with
						f s (n,2,True)= s+++n+++s
						f s (n,i,_) = n+++concat (repeatn i s)
				"Enum_"
					# list = [(n,i)\\(n,0,_)<-list & i<-[0..]]
					# list` = ["fromEnum_ "+++n+++" ="+++toString i+++"::Int_;"\\(n,i)<-list]
					# list`` = ["toEnum_ "+++toString i+++"="+++n+++";"\\(n,i)<-list]
					# max = toString (length list - 1)
					= list` ++ list`` ++
					 	[	"toEnum_ i=error_(\"fromEnum derived for "+++ tn +++" applied to \"++show_ i);"] ++
						[	"succ_ x=toEnum_(1+.fromEnum_ x);"
						,	"pred_ x=toEnum_(fromEnum_ x+.1);"
						,	"enumFrom_ x=map_ toEnum_[fromEnum_ x.."+++max+++"];"
						,	"enumFromTo_ x y=map_ toEnum_[fromEnum_ x..fromEnum_ y];"
						,	"enumFromThen_ x y=map_ toEnum_[fromEnum_ x,fromEnum_ y.."+++max+++"];"
						,	"enumFromThenTo_ x y z=map_ toEnum_[fromEnum_ x,fromEnum_ y..fromEnum_ z]"]

				"Read_"
					=	[	"readsPrec_=error_\"readsPrec derived for "+++ tn +++"\";"
						,	"readList_=error_\"readList derived for "+++ tn +++"\""]
				"Show_"
					=	[	"show_=error_\"show derived for "+++ tn +++"\";"
						,	"showsPrec_=error_\"showsPrec derived for "+++ tn +++"\";"
						,	"showList_=error_\"showList derived for "+++ tn +++"\""]
				"Ix_"
					=	[	"range_=error_\"range derived for "+++ tn +++"\";"
						,	"index_=error_\"index derived for "+++ tn +++"\";"
						,	"inRange_=error_\"inRange derived for "+++ tn +++"\";"
						,	"rangeSize_=error_\"rangeSize derived for "+++ tn +++"\""]

				_		= []

			# (instance_def,st)= case prg of
				[]		= ([], error (name+++" could't be derived") st)
				prg
						# inst= ["instance "]++ if (isEmpty args) [] (["("]++intersperse","[name+++" "+++a\\a<-args]++[")=>"]) ++ [name,"(",tn," "]++intersperse" "args++[")where{"]
						# prg= concat (inst++prg++["}"])
					 	= parse_tokens_with topdecl prg st

			= (instance_def,st)
	= ([PD_Type` type`:flatten decls], st)
where
	concat [] = ""
	concat [hd:tl] = hd+++.concat tl


tr_patt list=:(PE_List l) st
	= case l of
/*
f (a+1) = xxx
=>
f a` = xxx where a=a`-1
*/
			
		[id1=:PE_Ident a,PE_Ident b,num=:PE_List [PE_Ident _,PE_Basic (BVInt _)]] | toString b=="+.?"
			# (id,st)= get_new_id st
			# (id,st)= make_ident id st
			# (app,st)= make_application "-." [id,num] st
			# rhs= buildRhs app
			# d= PD_NodeDef NoPos id1 rhs
			= ([d],Nothing,id,st)
/*		[PE_Ident n,num=:PE_List [PE_Ident _,PE_Basic (BVInt _)]] | toString n=="-."
			# (id,st)= get_new_id st
			# (id,st)= make_ident id st
			# (num,st)= make_application "negate_" [num] st
			# (guard,st)= make_application "==." [id, num] st
			= ([],Just guard,id,st)
*/		[PE_Ident n,hd`:_] | is_appl n && not (is_conop hd`)
			# (id,st)= get_new_id st
			# (id,st)= make_ident id st
			# (guard,st)= make_application "==." [id, list] st
			= ([],Just guard,id,st)

		_
			# (le,gu,l,st)= tr_patts l st
			= (le,gu,PE_List l,st)
tr_patt	(PE_Bound b=:{bind_src}) st
	# (le,gu,bind_src,st)= tr_patt bind_src st
	= (le,gu,PE_Bound {b&bind_src= bind_src },st)
tr_patt (PE_Tuple l) st
	# (le,gu,l,st)= tr_patts l st
	= (le,gu,PE_Tuple l,st)
tr_patt expr st = ([],Nothing,expr,st)
	
tr_patts l st
	# (l,(le,gu,st))= maplSt f l ([],Nothing,st) with
		f p (le,gu,st)
			# (le2,gu2,p,st)= tr_patt p st
			# (gu,st)= merge_guard gu gu2 st
			= (p,(le++le2,gu,st))
	= (le,gu,l,st)
where
	merge_guard (Just a) (Just b) st
		# (c,st)= make_application "&&." [a, b] st
		= (Just c,st)
	merge_guard Nothing x st = (x,st)
	merge_guard x Nothing st = (x,st)

make_class pos context class_name args members st
	# class_args= map (\x=x.atv_variable) args
	# class_arity= length class_args

	# (members,st)= bind_types_and_bodies members st
	# class_def
	  	= { class_name = class_name, class_arity = class_arity, class_args = class_args,
	    	class_context = context, class_pos = pos, class_members = {}, class_cons_vars = 0 /*AN_Nones*/,
			class_dictionary = { ds_ident = { class_name & id_info = nilPtr }, ds_arity = 0, ds_index = NoIndex },
			class_arg_kinds = []}

	# (yes,st)= is_icl st
	# members= if yes members (filter is_ts  members)
	= ([PD_Class class_def members], st)

make_instance pos context class_id id types defs st
//	# defs= groupBy` pd_name defs
//	# (defs,st)= maplSt lift_function defs st
//	# defs= flatten defs
	# (defs,st)= bind_types_and_bodies defs st

	# (iclmodule,st)= is_icl st
	= [PD_Instance
			{	pi_class = class_id, pi_ident = id
			,	pi_types = types, pi_context = context
			,	pi_members = if iclmodule defs [], pi_specials = SP_None
			,	pi_pos = pos}] -- st

make_context (x,y) st
	# (x,st) = into_hashTable x IC_Class st
	# tc_class= TCClass {glob_object = MakeDefinedSymbol x NoIndex 1/*length types*/, glob_module = NoIndex}
	# (y,st)= into_hashTable y IC_Type st
	= {tc_class= tc_class, tc_var= nilPtr, tc_types= [TV (MakeTypeVar y)]} -- st

make_strictness_list types
	= add_strictness_for_arguments types 0 0 NotStrict
where 
	add_strictness_for_arguments :: ![Bool] !Int !Int !StrictnessList -> StrictnessList
	add_strictness_for_arguments [] strictness_index strictness strictness_list
		| strictness==0
			= strictness_list
			= append_strictness strictness strictness_list
	add_strictness_for_arguments [True:types] strictness_index strictness strictness_list
		# (strictness_index,strictness,strictness_list) = add_next_strict strictness_index strictness strictness_list
		= add_strictness_for_arguments types strictness_index strictness strictness_list
	add_strictness_for_arguments [False:types] strictness_index strictness strictness_list
		# (strictness_index,strictness,strictness_list) = add_next_not_strict strictness_index strictness strictness_list
		= add_strictness_for_arguments types strictness_index strictness strictness_list

make_tupleType [at_type] = at_type
make_tupleType types
	# arity= length types
	= TAS (MakeNewTypeSymbIdent predefined_idents.[GetTupleTypeIndex arity] arity) (map add_TA_None types) NotStrict

make_symbolType context result =
		{	st_vars = [], st_args = []
		,	st_args_strictness = NotStrict
		,	st_arity = 0, st_result = add_TA_None result
		,	st_context = context, st_attr_env = [], st_attr_vars = [] }

make_ewl pos expr=
			{	ewl_nodes	= []
			,	ewl_expr	= expr
			,	ewl_locals	= NoLocalParsedDefs//:: !LocalDefs //???
			,	ewl_position= pos
			}

tr_minus [x] = add x
tr_minus l = f True l .> app_val PE_List where
	f _ [x] st = ([x],st)
	f True [PE_Ident n:tl] st | toString n=="-.?"
		# (neg,st)= make_ident "negate_" st
		# (tl,st) = f False tl st
		# tl = case tl of
			[a] = a
			l	= PE_List l
    	= ([neg,tl], st)
	f _ [hd:tl] st
		# (tl,st) = f (isOp hd) tl st
		= ([hd:tl],st)

make_lambda ident args st
	# (le,gu,args,st)= tr_patts args st
	| isJust gu || not (isEmpty le) = (dummy, error "guard ignored" st)

	# (pos,st)= getPos st
	# st= want_token Arrow_Token st
	# (result,st)= (try_exp !. "exp") st
	= (PE_Lambda ident args result pos, st)

make_varsym name st
//		| has_suffix
	# name= if (size name>1 && name.[size name-1]=='.') (name+++"?") name
	# (name,st)= make_ident name st
	= (name, st)

make_bracketed_expr l st = case l of
	[a]	| isSymOp a = (PE_List l,st)
	[PE_List l]	| length l>=2 && (isOp (hd l) || isOp (last l))
		# (lam_ident,st)= internalIdent "\\" st
		# (arg,st)= get_new_id st
		# (arg,st)= make_ident arg st 
		# (pos,st)= getPos st
		# result= if (isOp (hd l))
			(PE_List [arg,hd l,case tl l of [i]=i; i=PE_List i])
			(PE_List [case init l of [i]=i; i=PE_List i,last l,arg])
		= (PE_Lambda lam_ident [arg] result pos, st)
	[x]				= (x,st)
	l				= (PE_Tuple l,st)

make_tuple_fun num st
	# (lam_ident,st)= internalIdent "\\" st
	# (pos,st)= getPos st
	# (args,st)= get_new_ids num st
	# (args,st)= maplSt make_ident args st
	= (PE_Lambda lam_ident args (PE_Tuple args) pos, st)

make_module module_name modification_time mod st
	// automatically import the preludes
	# (prelude_name,st)= into_hashTable (if (module_name=="Prelude_") "PreludeBuiltin" "Prelude_") IC_Module st
	# impdecl= PD_Import [makeParsedImport NoPos prelude_name []]
	# mod= [impdecl: mod]

	# (module_ident, st) = into_hashTable module_name IC_Module st
	=
		{ mod_name 				= module_ident
		, mod_modification_time = modification_time
		, mod_type 				= MK_Module
		, mod_imports 			= []
		, mod_imported_objects 	= []
		, mod_defs 				= mod
		} -- st

is_declaration (PD_Function _ _ _ _ _ FK_Macro) = True
is_declaration (PD_Function _ _ _ _ _ _) = False
//is_declaration (PD_TypeSpec _ _ _ No _) = False
is_declaration _ = True

resort_dcl l
	# (a,b)= partition p l
	= a++b
where
	p (PD_TypeSpec _ _ _ (Yes {st_context}) _) = not (isEmpty st_context)
	p _ = False

make_typedef pos name context vars rhs
	= {	td_name = name, td_index = -1, td_arity = length vars, td_args = vars, 
		td_attrs = [], td_attribute = TA_None, td_context = context,
		td_pos = pos, td_rhs = rhs, td_used_types = [] }

PD_Type` {td_name} | toString td_name=="Bool_"	= PD_Import []
PD_Type` t = PD_Type t

make_cons_def pos name prio types strict=
		{	pc_cons_name = name, pc_arg_types = map add_TA_None types, 
			pc_args_strictness= make_strictness_list strict, pc_cons_arity = length types,
	  		pc_cons_prio = prio, pc_exi_vars = [], pc_cons_pos = pos}


into_ht t (s, st) = into_hashTable s t st
into_ht` t s st = into_hashTable s t st

make_alt pos pattern expr locals st
	# ewl= make_ewl pos expr
	# (pattern,rhs,st)= make_rhs` [pattern] (UnGuardedExpr ewl) locals st
	# case_alt = { calt_pattern = make_exp pattern, calt_rhs = rhs }
	= (case_alt, st)

comp Nothing x st = (x,st)
comp (Just guard) x st
	# ({lc_line},st)= getLineAndColumn st
	# guard=	{ alt_nodes = [], alt_guard = guard, alt_expr = x,
			  alt_ident = guard_ident , alt_position = NoPos } with
					guard_ident
						= { id_name = "_g;" +++ toString lc_line +++ ";", id_info = nilPtr }
	= GuardedAlts [guard] No -- st

buildRhs` le guard rhs st
	# alts = UnGuardedExpr (make_ewl NoPos rhs)
	# (alts,st)= comp guard alts st
	# rhs = {rhs_alts= alts, rhs_locals= LocalParsedDefs le}
	= (rhs,st)

make_rhs` args rhs locals st
	# (le,gu,args,st)= tr_patts args st
	# (rhs,st)= comp gu rhs st
	# rhs= {rhs_alts= rhs, rhs_locals= add_to_locals le locals}
	= (args,rhs,st)
where
	add_to_locals le (LocalParsedDefs x)= LocalParsedDefs (le++x)
	add_to_locals _ _ = abort "add_to_locals"

make_rhs pos op args rhs locals st
	# (args,rhs,st)= make_rhs` args rhs locals st
	= (f pos op args rhs, st)
where
	f pos op args rhs
		| is_ok op	= [PD_Function pos op False args rhs (FK_Fun)]
		= [PD_NodeDef pos (make_exp args) rhs]

make_type_inside pos pos2 exp type st
	# (v_ident,st)= internalIdent` "v" pos2 st
	= PE_Let False
		(LocalParsedDefs
			[	PD_TypeSpec pos v_ident NoPrio type SP_None
			,	PD_Function pos v_ident False [] (buildRhs exp) (FK_Fun)
			])
		(PE_Ident v_ident) -- st



///////////////////////////////////////////////////////////////////////////////////
// Parsing
//
		
want_module module_name modification_time =
	(iff` Module_Token
		(	want_modid ..> \modid=
			if_ (modid<>module_name) (expected module_name) .>
			want_exports ..> \exports=
			want_token Where_Token .>
			want_body exports)

		(	into_ht` IC_Expression "main" ..> \main=
			want_body (Yes [main]))
	) ..> make_module module_name modification_time

want_body exports =
	want_token CurlyOpen_Token .>
	impdecls_and_topdecls .>
	uncurry bind_types_and_bodies .>
	uncurry add_instance_defs .>
	uncurry bind .$
	CurlyClose_Token
where
	bind decls = is_icl ?> snd .> add decls <|> add (resort_dcl (filter is_declaration decls))


impdecls_and_topdecls = 
		try_impdecl ?.> (\impdecl=
			Semicolon_Token $. impdecls_and_topdecls ..> (\tl = add [impdecl:tl])
		<|>	add [impdecl])
	<|> topdecl ?.> (\topdecl`=
 			Semicolon_Token $. listWith topdecl Semicolon_Token ..> (\decls= add (flatten [topdecl`:decls]))
		<|>	add topdecl`)
	<|>	add []


want_exports =  Open_Token $. commaList want_export .$ Close_Token   <|> add []

want_export =
		Module_Token $. want_modid
	<|>	qtycon ?> snd .> (
			Open_Token $. (
				DotDot_Token $. add ".."
			<|>	commaList (qcname !. "qcname") ?.> \_= add " "
					) .$ Close_Token
		<|> add " ")
	<|> protect qvar
	<|> Open_Token $. consym !. "varsym" .$ Close_Token  //Hugs

qcname = protect qvar <|> qcon

try_impdecl =  Import_Token $. want_impdecl

want_impdecl =
	getPos ..> \pos=
	try_token (Var_Id "qualified_") ..> \yes=
	if_ yes (warning`` "keyword qualified ignored") .>
	want_modid .> into_ht IC_Module ..> \modid=
//	# (pos`,st)= getLineAndColumn st
	want_impspec ..> \items=
	add (PD_Import [makeParsedImport pos modid []/*items*/])

want_impspec = 
		 (Var_Id "hiding_") $. (want_token Open_Token .> commaList want_import .$ Close_Token)
	<|>  Open_Token $. (commaList want_import .$ Close_Token)

want_import =
		var ?> into_ht IC_Expression ..> (\id = add (ID_Function {ii_ident = id, ii_extended = False }))
	<|> tycon !. "ident" ?> snd .> add (ID_OldSyntax [])

want_cname = var <|> want_con

make_fields _ st = ([],st)

topdecl :: (*(State .a) -> *(![ParsedDefinition],!*(State .a)))
topdecl =
	getPos ..> \pos=
		Type_Token 	$. (
		simpletype ..> \(name,vars)=
		want_token Equal_Token .>
		want_type ..> \type=
		add [PD_Type (make_typedef pos name [] vars (TypeSpec (add_TA_None type)))]
	 	) 
	<|> Data_Token 	$. (
		try_scontext_and_arrow ..> \context=
		simpletype ..> \(name,vars)=
		iff` Equal_Token
		(	listWith constr Bar_Token ..> \list=
			let type= make_typedef pos name context vars (ConsList list)//(map snd list))
			in deriv type /* ..> \defs=
			make_fields list ..> \defs`=
			add (defs`++defs)*/
		)
		$	add [PD_Import []]
		)
	<|> Newtype_Token 	$. (
		try_scontext_and_arrow ..> \context=
		simpletype ..> \(name,vars)=
		want_token Equal_Token .>
		newconstr ..> \type=
		deriv (make_typedef pos name context vars (ConsList [type]))
		)
	<|>	Class_Token 	$. (
		try_scontext_and_arrow ..> \context= 
		(tycls !. "class name" .> into_ht IC_Class) ..> \name=
		tyvar !. "type variable" ..> \arg=
		want_token Where_Token .>
		want_cdecls ..>
		make_class pos context name [arg]
		)
	<|> Instance_Token $. (
		try_scontext_and_arrow ..> \context=
		qtycls !. "class name" ..> \name=
		into_ht` IC_Class name ..> \id1=
		inst ..> \type= 
		into_ht` (IC_Instance [type]) name ..> \id=
		want_token Where_Token .>
		want_idecls ..>
		make_instance pos context id1 id [type]
		)
	<|> (Var_Id "primitive_") $. (
		want_var .> snd .>
		iff` Comma_Token (commaList want_var .> snd) (try_token_by get_string .> snd) .>
		want_token DoubleColon_Token .>
		context_and_type .> snd .>
		add [PD_Import []]
		)
	<|>  try_decl
where
	get_string (String_Token s) 	= Yes (BVS s)
	get_string _					= nothing


want_decls = 
	(groupWith try_decl .>
	app_val flatten ..> 
	bind_types_and_bodies) .> 
	app_val LocalParsedDefs

try_decl =
		try_gendecl
	<|> protect (try_pat`` ?.> try_rhs`)
	<|> try_funlhs ?.> try_rhs`

try_pat`` = try_pat` False ?> app_val f where
	f 		(PE_Ident a) = (a, Yes [])
	f		pat 		= (nothing, Yes [pat])

want_cdecls = groupWith want_cdecl .> app_val flatten
want_cdecl = 
		try_gendecl  
	<|> protect (var ?> into_ht IC_Expression ?> app_val (\op=(op,Yes [])) ?.> try_rhs`) 
	<|> try_funlhs !. "lhs" ..> try_rhs`

want_idecls = groupWith want_idecl .> app_val flatten
want_idecl = want_cdecl //try_gendecl  <|> (try_funlhs ?.> want_rhs```

//try_gendecl :: (*(State .a) -> *(!Optional [ParsedDefinition],!*(State .a)))
try_gendecl = try_fixitydec <|> try_signature

try_fixity = 
		 Infix_Token $. add (Yes NoAssoc)
	<|>	 Infixl_Token $. add (Yes LeftAssoc)
	<|>	 Infixr_Token $. add (Yes RightAssoc)

want_ops = commaList want_op

//try_fixitydec :: *(State .a) -> *(!Optional [ParsedDefinition],!*(State .a))
try_fixitydec =
	try_fixity ?.> \(Yes assoc)= 
	want_digit ..> \digit= 
	make_ident "a" ..> \a=
	make_ident "b" ..> \b=
	commaList (
		getPos ..> \pos= want_op ..> make_fixitydef pos (Prio assoc digit) [a,b]
	) .> app_val flatten

want_digit = (integer <|> add (Yes 9)) ..> \(Yes n) st = if (n>=0 && n<=10) (n,st) (9,expected "digit" st)

want_type = btype ..> \type=
		Arrow_Token $. want_type ..> (\types= add (add_TA_None type --> add_TA_None types))
	<|>	add type

btype = oneOrMore atype "atype3" ..> make_btype

atype =
		qtycon ?> into_ht IC_Type .> app_val (\name= TA (MakeNewTypeSymbIdent name 0) [])
	<|>	tyvar` ?> app_val TV
	<|>	Open_Token $. commaList want_type .> app_val make_tupleType .$ Close_Token
	<|>	SquareOpen_Token $. (
			SquareClose_Token $. add (makeListTypeSymbol [])
		<|>	want_type ..> \type= add (makeListTypeSymbol [add_TA_None type]) .$ SquareClose_Token
		)

try_scontext_and_arrow =
//	try_scontext ?> app_st (want_token DoubleArrow_Token)  <|> add []
//try_scontext = 
	(protect ((
			Open_Token $. commaList` try_class ?> app_st (want_token Close_Token)
		<|> try_class ?> app_val \x=[x]
			) ?> add_val` (try_token DoubleArrow_Token))
		) ..> maplSt make_context
where
	try_class = conid ?.> \x= varid ?.> \y= add (x,y)

context_and_type =
	try_scontext_and_arrow ..> \context=
	want_type ..> \result=
	add (Yes (make_symbolType context result))
	
simpleclass = qtycls !. "class name1" .> into_ht IC_Class ..> \name= tyvar !. "type variable" ..> \v= add (name,v)

simpletype =
	tycon !. "tycon" .> into_ht IC_Type ..> \name=
	zeroOrMore tyvar ..> \vars=
	add (name,vars)

deriv type =
	iff` Deriving_Token
		((	Open_Token $. commaList p .$ Close_Token
		<|> p .> app_val \x=[x]) ..> make_deriv type)
		(add [PD_Type` type])
where
	p	= conid !. "class name4" ..> \n= getLineAndColumn ..> \pos= add (n,pos)

newconstr =
	getPos ..> \pos=
	conid !. "constructor" .> into_ht IC_Expression ..> \name`=
	zeroOrMore atype ..> \types=
	add (make_cons_def pos name` NoPrio types [])

fielddecl = 
	var !. "var" ..> \name=
	want_token DoubleColon_Token .> (
		 		(Var_Sym "!.") $. atype !. "atype" ..> (\t= add (True,t))
			<|>	want_type ..> (\t= add (False,t))
			) ..> \type=
	add type //!!!

//							# (field_name, st)= into_hashTable field_n (IC_Field type_name) st
//							# (selector_name, st)= into_hashTable field_n IC_Selector st
//							# (field_var, st)= into_hashTable field_n IC_Expression st

atype` =
 		(Var_Sym "!.") $. atype !. "atype1" ..> (\t= add (True,t))
	<|>	atype ..> (\t= add (False,t))

btype` =
 		(Var_Sym "!.") $. atype !. "atype2" ..> (\t= add (True,t))
	<|>	btype ..> (\t= add (False,t))

constr =
	getPos ..> \pos=
	(			protect (btype` .> add_val (want_conop.> into_ht IC_Expression)) ?.> (\(type1,op)= btype` ..> \type2= add (op,Prio LeftAssoc 9,[type1,type2]))
			<|>	conid !. "constructor" .> into_ht IC_Expression ..> \name=
				(	CurlyOpen_Token $. commaList fielddecl .$ CurlyClose_Token
				<|>	zeroOrMore atype`) ..> \types=
						add (name,NoPrio,types)
	) ..> \(name,prio,types)=
	let (strict,types`) = unzip types 
	in add (make_cons_def pos name prio types` strict)

inst = atype !. "inst"

//want_type :: !*(State .a) -> *(!Type,!*(State .a))	 


//try_signature :: *(State .a) -> *(!Optional [ParsedDefinition],!*(State .a))
try_signature =
	getPos ..> \pos= 
	pp ?.> \vars= maplSt (into_ht` IC_Expression) vars ..> \vars= 
	context_and_type ..> \type=
	add [PD_TypeSpec pos name NoPrio type SP_None \\ name<- vars]
where
	pp = protect (commaList` var ?> add_val` (try_token DoubleColon_Token))

try_funlhs =  
		protect (try_pat` False ?> add_val want_varop) ?.>
			 (\(p1,op)= try_pat True !. "pat" ..> \p2= curry (into_ht IC_Expression) op ..> \op= add (op,Yes [p1,p2]))
	<|> try_other

//try_other = (Open_Token $. varsym !. "varsym" .$ Close_Token  <|>  varid) ?> into_ht IC_Expression ..> \op= zeroOrMore (try_apat True) ..> \args= add (op,args)
try_other = (
		Open_Token $. try_funlhs !. "lhs" .$ Close_Token  
	<|>	varid ?> into_ht IC_Expression ..> \op= add (op,Yes [])
	) ?.> \(op,Yes args)= zeroOrMore (try_apat True) ..> \args`= add (op,Yes (args++args`))

try_rhs` :: (Ident,Optional [ParsedExpr]) -> (*(State .a) -> *(![ParsedDefinition], !*(State .a)))
try_rhs` (op,Yes args) =
	function_body ?.> \(Yes rhs)= 
	getPos ..> \pos=
	(Where_Token $. want_decls) ..> \locals=  //!
	make_rhs pos op args rhs locals

function_body =
		guard ?.> (\hd= zeroOrMore guard ..> \tl= add (Yes (GuardedAlts [g\\Yes g<-[hd:tl]] No)))
	<|>	Equal_Token $. try_exp !. "exp" .> app_val (Yes o UnGuardedExpr o make_ewl NoPos)

guard = 
	Bar_Token $. getLineAndColumn ..> \{lc_line}= 
	try_exp !. "exp" ..> \guard=
	want_token Equal_Token .> 
	try_exp !. "exp" .> app_val (UnGuardedExpr o make_ewl NoPos) ..> \body=
	add	(Yes { alt_nodes = [], alt_guard = guard, alt_expr = body
		, alt_ident= { id_name = "_g;" +++ toString lc_line +++ ";", id_info = nilPtr } , alt_position = NoPos })


try_exp = 
	try_aexp ?.> \hd=
		zeroOrMore try_aexp ..> \tl= 
		tr_minus [hd:tl] ..> \exp=
		try_token DoubleColon_Token ?> snd .> (
			getPos ..> \pos=
			getLineAndColumn ..> \pos2=
			context_and_type ..> \type=
			make_type_inside pos pos2 exp type
			)
		<|>	add exp

try_aexp :: (*(State .a) -> *(!ParsedExpr,*(State .a)))
try_aexp =
	 	qvarid ?.> make_ident
	<|> qvarsym ?.> make_varsym

	<|> (conid ?.> make_ident) ?> app_st
			(iff` CurlyOpen_Token
			/*commaList_or_nothing` p "ssss" .> */ 
				(warning`` "{} ignored" .> want_token CurlyClose_Token)
				id
			)
	<|>	consym ?.> make_ident

	<|>	literal

	<|> Prime_Token $. (
		varid !. "varid" ..> \var= 
		make_ident (var+++"9") .>
		app_st (want_token Prime_Token))

	<|>	Open_Token $. (
			zeroOrMore (try_token Comma_Token) ?.> (\l= make_tuple_fun (length l+1))
		<|> commaList (try_exp !. "exp") ..> (\l= make_bracketed_expr l)
		) .$ Close_Token

	<|>	SquareOpen_Token $. listExpr .$ SquareClose_Token

	<|> BackSlash_Token $. (
		internalIdent "\\" ..> \ident=
		oneOrMore (try_apat True) "pattern" ..> \args=
		make_lambda ident args)
	<|> Let_Token $. (
		want_decls ..> \defs= 
		want_token In_Token .> 
		(try_exp !. "exp") ..> \exp= 
		add (PE_Let False defs exp))
	<|> If_Token $. (
		internalIdent "_if" ..> \ident= 
		(try_exp !. "exp") ..> \cond= 
		want_token Then_Token .>
		(try_exp !. "exp") ..> \then=
		want_token Else_Token .>
		(try_exp !. "exp") ..> \else=
		add (PE_If ident cond then else))
	<|> Case_Token $. (
		internalIdent "_c" ..> \ident=
		try_exp !. "exp" ..> \exp=
		want_token Of_Token .>
		groupWith want_alt ..> \alts=
		add (PE_Case ident exp alts))
	<|> Do_Token $. groupWith try_qual ..> transform_do

//want_qual :: *(State .a) -> (!HQualifier,!*(State .a))	 
try_qual = 
		iff` Let_Token
			(want_decls .> app_val HLocal)
	$	iff (protect (try_pat` True .> add_val` (try_token LeftArrow_Token)))
	  (\patt=
		getPos ..> \pos=
		getLineAndColumn ..> \pos2=
		try_exp !. "exp" ..> \exp=
		add (HGenerator pos pos2 patt exp))
		(try_exp !. "exp" .> app_val HGuard)

want_alt =
	try_pat True !. "pat" ..> \pattern=
	want_token Arrow_Token .>
	getPos ..> \pos=
	try_exp !. "exp" ..> \expr=
	(Where_Token $. want_decls) ..>
	make_alt pos pattern expr //locals

listExpr =
		try_exp ?.> (\e1=
			Comma_Token $. try_exp !. "exp" ..> (\e2=
				Comma_Token $. commaList (try_exp !. "exp") ..> (\tl= make_list [e1,e2:tl])
			<|>	DotDot_Token $. (
					try_exp ?.> (\e3= make_application "enumFromThenTo_" [e1,e2,e3])
				<|> make_application "enumFromThen_" [e1,e2]
				)
			<|>	make_list [e1,e2]
			)
		<|>	DotDot_Token $. (
				try_exp ?.> (\e2= make_application "enumFromTo_" [e1,e2])
			<|> make_application "enumFrom_" [e1]
			)
		<|>	Bar_Token $. commaList try_qual ..> transform_qualifier e1
		<|>	make_list [e1]
		)
	<|>	make_list []

try_pat must =
	(	(Var_Sym "-.") $. make_ident "negate_"
	<|>	try_apat must) ?.> \hd=
			zeroOrMore (try_apat must) ..> \tl= add (make_exp [hd:tl])

//try_apat :: Bool *(State .a) -> *(!Optional ParsedExpr,*(State .a))
try_apat must =
		Open_Token $. commaList (try_pat True !. "pat") .> app_val make_tuple .$ Close_Token
	<|>	other_apat must
	
other_apat must =
		SquareOpen_Token $. want_listpat .$ SquareClose_Token
	<|>	consym ?.> make_ident
	<|>	literal
	<|> (conid ?.> make_ident) ?> app_st (
			iff` CurlyOpen_Token 
				(commaList_or_nothing` p "ssss" .> snd .> warning`` "{} ignored" .> want_token CurlyClose_Token)
				id
			)
	<|> varid ?> into_ht IC_Expression ..> (\id=
			At_Token $. try_apat True !. "try_aexp" ..> (\pat= add (PE_Bound { bind_dst = id, bind_src = pat }))
		<|> add (PE_Ident id))
	<|> WildCard_Token $. add PE_WildCard
	<|> Tilde_Token $. warning`` "~ ignored" .> try_apat True
	<|> add must ?> snd .> (Var_Sym "+.") $. make_ident "+.?"

where
	p = varid ?.> \varid= want_token Equal_Token .> try_pat True !. "pat" ..> \pat= add (varid,pat)

//try_pat` :: Bool *(State .a) -> *(!Optional ParsedExpr,*(State .a))
try_pat` must =
	(	(Var_Sym "-.") $. make_ident "negate_"
	<|>	try_apat` must) ?.> \hd=
			iff (protect (try_apat must))
				(\hd`=
				add (is_con hd || is_conop hd`) ?> snd .>
					zeroOrMore (try_apat must) ..> \tl=
					add (PE_List [hd,hd`:tl])
				)
				(add hd)

//fff (a-- b) = a -- b

try_apat` must =
		Open_Token $. (try_pat` True ?.> \hd= 
			Comma_Token $. commaList (try_pat True !. "pat") ..> (\tl= add [hd:tl])
		<|> add [hd]
		)  ?> app_val make_tuple ?> add_val` (try_token Close_Token)
	<|>	other_apat must

want_listpat = commaList_or_nothing` (try_pat True) "pat" ..> make_list


var 		= varid 	<|>	Open_Token $. varsym ?> add_val` (try_token Close_Token)
want_var 	= var !. "var"
qvar 		= qvarid 	<|>	Open_Token $. qvarsym ?> add_val` (try_token Close_Token)
want_qvar 	= qvar !. "qvar"
con 		= conid 	<|>	Open_Token $. consym ?> add_val` (try_token Close_Token)
want_con 	= con !. "con"
qcon 		= qconid 	<|>	Open_Token $. gconsym ?> add_val` (try_token Close_Token)
want_qcon 	= qcon !. "qcon"
want_varop 	= varsym 	<|>	 Prime_Token $. varid !. "varid" .$ Prime_Token
want_qvarop = qvarsym 	<|>	 Prime_Token $. qvarid !. "varid" .$ Prime_Token
want_conop 	= consym  	<|>	 Prime_Token $. conid !. "conid" .$ Prime_Token
want_qconop = qconsym  	<|>	 Prime_Token $. qconid !. "conid" .$ Prime_Token
want_op 	= want_varop <|> want_conop
want_qop 	= want_qvarop <|> want_qconop
gconsym		= qconsym	//!!!

/////////////

tyvar` 		= varid ?> into_ht IC_Type .> app_val MakeTypeVar
tyvar 		= tyvar` ..> \var= add {atv_attribute = TA_None, atv_variable = var}
tycon		= conid
tycls		= conid
modid		= conid //@@ app_val transform_id
want_modid	= modid !. "modid"

qvarid		= qualified varid
qconid		= qualified conid
qtycon		= qualified tycon
qtycls		= qualified tycls
qvarsym		= qualified varsym
qconsym		= qualified consym

qualified p = p // <|> protect (modid ?> snd .> (Var_Sym "..") $. p .> app_st (warning "Qualified name"))

varid 		= try_token_by get_varId where
	get_varId (Var_Id x) 	= x
	get_varId _ 			= nothing

varsym 		= try_token_by get_varSym where
	get_varSym (Var_Sym x) 	= x
	get_varSym _ 			= nothing

conid 		= try_token_by get_conId where
	get_conId (Con_Id x) 	= x
	get_conId _ 			= nothing

consym 		= try_token_by get_conSym where
	get_conSym (Con_Sym x) 	= x
	get_conSym _ 			= nothing

integer 	= try_token_by f where
	f (Int_Token x) 		= Yes x
	f _ 					= nothing

float 		= try_token_by f where
	f (Float_Token a b) 		= Yes (a,b)
	f _ 					= nothing

char	 	= try_token_by f where
	f (Char_Token x) 		= x
	f _ 					= nothing

string` 	= try_token_by f where
	f (String_Token x)	 	= x
	f _ 					= nothing

literal 	=
		integer ?> app_val (\(Yes n)= (PE_Basic o BVInt) n) ..> (\x= make_application "fromInteger_" [x])
	<|>	float ?.> make_float
	<|> string` ?> app_val (PE_Basic o BVS) ..> (\x = make_application "string_to_charlist_" [x])
	<|> char ?> app_val (PE_Basic o BVC)

make_float (Yes (n,m)) st
	# (e,st)= make_application "%." [PE_Basic (BVInt n),PE_Basic (BVInt m)] st
	= make_application "fromRational_" [e] st



//////////////////////////////////////////////////////////////
// The definition of the wantModule function
//


wantModule :: !Bool !Ident !Position !Bool !*HashTable !*File !SearchPaths (ModTimeFunction *Files) !*Files
	-> (!Bool, !ParsedModule, !*HashTable, !*File, !*Files)
wantModule iclmodule file_id=:{id_name} import_file_position support_generics hash_table error searchPaths modtimefunction files
	| size id_name>1 && id_name.[size id_name-1]=='_'  // this is a Haskell module
		# file_name= id_name+++".dcl"
		= case fopenInSearchPaths iclmodule id_name searchPaths FReadData modtimefunction files of
			(No, files)
				-> let mod = { mod_name = file_id,  mod_modification_time = "", mod_type = MK_None, mod_imports = [], mod_imported_objects = [], mod_defs = [] } in
				  (False, mod, hash_table, error <<< "Error " <<< import_file_position <<< ": "  <<< file_name <<< " could not be imported\n", files)
			(Yes (file, modification_time), files)
				# hash_table=set_hte_mark (if iclmodule 1 0) hash_table
				# st= open_state file file_name iclmodule 8/*default tabsize in Haskell*/ error hash_table
				# (mod,st)= want_module id_name modification_time st
				# (ok, file, error, hash_table)= close_state st
				# files= snd (fclose file files)
				# hash_table=set_hte_mark 0 hash_table
				->(ok,mod,hash_table,error,files)

	= wantModule_original iclmodule file_id import_file_position support_generics hash_table error searchPaths modtimefunction files

//try_to_open_InSearchPaths :: ![String] SearchPaths !Int (ModTimeFunction *f) !*f -> (Optional (*File, String, {#Char}),!*f) | FileSystem f
fopenInSearchPaths iclmodule modname searchPaths mode modtimefunction f
	= case [ path \\ (name,path)<- searchPaths.sp_locations | name==modname+++".dcl" ] of
		[path:_]	= try_to_open path f
		[]			= fopenAnywhereInSearchPaths searchPaths.sp_paths f
where
	try_to_open path f
		# fileName	= path+++modname+++".dcl"
		# (opened, file, f) = fopen fileName mode f
		| opened
			# (time, f)	= modtimefunction (path+++modname+++if iclmodule ".icl" ".dcl") f
			| time<>""
				= (Yes (file,time), f)
			# (_, file`, f)	= fopen (path+++modname+++".icl") FWriteText f
			# (_,f)= fclose file` f
			# (time, f)	= modtimefunction (path+++modname+++if iclmodule ".icl" ".dcl") f
			= (Yes (file, time), f)
		= (No, f)

	fopenAnywhereInSearchPaths [] f		=	(No, f)
	fopenAnywhereInSearchPaths [path : paths] f
			= case try_to_open path f of
				(No, f)	= fopenAnywhereInSearchPaths paths f
				yes		= yes

